# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from .deconvolution import (
    create_linear_system,
    solve_linear_system,
    em_deconvolve,
    deconvolve,
    DeconvolutionResult,
)
from .design import Design
from .initialization import init
from .main import find_best_design, best_design_for_pool_budget
from .optimization import optimize
from .simulation import (
    simulate_elispot_counts,
    simulate_any_hits_per_pool,
    simulate_number_hits_per_pool,
)
from .types import SpotCounts
from .validity import is_valid, count_violations, violations_per_replicate

__version__ = "2.5.3"

__all__ = [
    "__version__",
    "find_best_design",
    "best_design_for_pool_budget",
    "Design",
    "init",
    "optimize",
    "count_violations",
    "is_valid",
    "violations_per_replicate",
    "simulate_elispot_counts",
    "simulate_any_hits_per_pool",
    "simulate_number_hits_per_pool",
    "create_linear_system",
    "solve_linear_system",
    "SpotCounts",
    "deconvolve",
    "em_deconvolve",
    "DeconvolutionResult",
]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np

from .design import Design
from .types import Replicate, Peptide
from .util import pairs_to_dict, transitive_closure


def _make_peptide_to_invalid_excluding_replicate(
    s: Design,
) -> dict[Replicate, dict[Peptide, set[Peptide]]]:
    peptide_to_invalid = pairs_to_dict(s.invalid_neighbors)
    peptide_to_invalid_excluding_replicate = {
        replicate_idx: {
            peptide: peptide_to_invalid.get(peptide, set()).copy()
            for peptide in range(s.num_peptides)
        }
        for replicate_idx in range(s.num_replicates)
    }
    replicate_to_peptide_to_pool_idx = s.replicate_to_peptide_to_pool_dict()

    for replicate_idx, pool_to_peptides_1 in s.assignments.items():
        for other_replicate_idx, pool_to_peptides_2 in s.assignments.items():
            if replicate_idx == other_replicate_idx:
                continue
            for peptides in pool_to_peptides_1.values():
                for peptide in peptides:
                    other_pool_idx = replicate_to_peptide_to_pool_idx[
                        other_replicate_idx
                    ][peptide]

                    for other_peptide in pool_to_peptides_2[other_pool_idx]:
                        if other_peptide != peptide:
                            peptide_to_invalid_excluding_replicate[replicate_idx][
                                peptide
                            ].add(other_peptide)
                            peptide_to_invalid_excluding_replicate[replicate_idx][
                                other_peptide
                            ].add(peptide)
    return peptide_to_invalid_excluding_replicate


def merge_small_pools(s: Design, verbose: bool = False) -> int:
    num_merged = 0

    peptide_to_preferred = transitive_closure(pairs_to_dict(s.preferred_neighbors))
    peptide_to_invalid_excluding_replicate = (
        _make_peptide_to_invalid_excluding_replicate(s)
    )

    replicate_order = list(range(s.num_replicates))

    for i in replicate_order:
        pool_to_peptides = s.assignments[i]

        merged = set()
        # next, try to merge any pools that are smaller than the max size
        for pool_idx_1, peptides_1 in list(pool_to_peptides.items()):
            if pool_idx_1 in merged:
                continue

            if len(peptides_1) >= s.max_peptides_per_pool:
                continue

            n_peptides_1 = len(peptides_1)
            if n_peptides_1 == 0:
                # skip empty pools, they'll be removed at the end
                continue

            candidates = list(pool_to_peptides.items())
            # split the candidates into those with preferred peptide overlaps and those without
            # while also filtering out pools that are invalid or have already been merged
            candidates_with_preferred_neighbors = []
            candidates_without_preferred_neighbors = []
            for pool_idx_2, peptides_2 in candidates:
                if pool_idx_1 == pool_idx_2:
                    # can't merge a pool with itself
                    continue
                if len(peptides_2) == 0:
                    # skip empty pools, they'll be removed at the end
                    continue
                if n_peptides_1 + len(peptides_2) >= s.max_peptides_per_pool:
                    # can't exceed the max pool size
                    continue
                if pool_idx_2 in merged:
                    continue

                any_preferred = False
                peptides_2_set = set(peptides_2)

                for p1 in peptides_1:
                    preferred = peptide_to_preferred.get(p1)
                    if not preferred:
                        continue
                    intersection = peptides_2_set.intersection(preferred)
                    if len(intersection) > 0:
                        any_preferred = True
                        break
                if any_preferred:
                    candidates_with_preferred_neighbors.append((pool_idx_2, peptides_2))
                else:
                    candidates_without_preferred_neighbors.append(
                        (pool_idx_2, peptides_2)
                    )

            for pool_idx_2, peptides_2 in (
                candidates_with_preferred_neighbors
                + candidates_without_preferred_neighbors
            ):
                all_valid = True
                # check if any peptides in pool 1 are invalid with any peptides in pool 2
                combined_peptides = list(peptides_1) + list(peptides_2)
                for peptide in list(combined_peptides):
                    other_peptides = {p for p in combined_peptides if p != peptide}
                    all_valid = (
                        len(
                            peptide_to_invalid_excluding_replicate[i][
                                peptide
                            ].intersection(other_peptides)
                        )
                        == 0
                    )
                    if not all_valid:
                        break

                if all_valid:
                    num_merged += 1
                    if verbose:
                        print(
                            "-- merging pools %d and %d in replicate %d"
                            % (pool_idx_1, pool_idx_2, i + 1)
                        )
                    pool_to_peptides[pool_idx_1] = np.array(combined_peptides)
                    pool_to_peptides[pool_idx_2] = np.array([])
                    merged.update([pool_idx_1, pool_idx_2])
                    for p in combined_peptides:
                        other_peptides = {p for p in combined_peptides if p != peptide}
                        for other_replicate_idx in range(s.num_replicates):
                            if other_replicate_idx != i:
                                peptide_to_invalid_excluding_replicate[
                                    other_replicate_idx
                                ][p].update(other_peptides)
                    break

    # just in case we ended up with any empty pools, remove them from the solution
    s.remove_empty_pools()
    return num_merged


def cleanup(s: Design, verbose: bool = True, max_iters=1000) -> int:
    total_num_merged = 0
    prev_num_pools = s.num_pools()

    for iter_idx in range(max_iters):
        num_merged = merge_small_pools(s)
        total_num_merged += num_merged
        num_pools = s.num_pools()
        if verbose:
            print(
                "Merge iter %d: merged %d small pools, final number of pools: %d (prev %d) "
                % (iter_idx + 1, num_merged, s.num_pools(), prev_num_pools)
            )
        if num_pools >= prev_num_pools:
            break
        prev_num_pools = num_pools
    return total_num_merged



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from .deconvolution import create_linear_system, solve_linear_system
from .design import Design
from .simulation import simulate_number_hits_per_pool
from .validity import count_violations


@dataclass
class EvaluationResult:
    precision: float
    recall: float
    f1: float
    num_pools: int
    num_violations: int

    def sort_key(self):
        # maximize f1, minimize violations,  maximize (precision, recall), minimize pools, maximize replicates
        return (
            -round(self.f1, 2),
            self.num_violations,
            -round(self.precision, 2),
            -round(self.recall, 2),
            self.num_pools,
        )


def evaluate_design(
    s: Design,
    num_simulation_iters: int = 10,
    min_hit_fraction: float = 0.01,
    max_hit_fraction: float = 0.05,
) -> EvaluationResult:
    """
    Returns average precision, recall, F1 scores across multiple simulations of the given design for
    num_hits in range of 1% to 5% of the total number of peptides
    """
    ps = []
    rs = []
    f1s = []
    for _ in range(num_simulation_iters):
        for hit_fraction in np.arange(min_hit_fraction, max_hit_fraction + 0.001, 0.01):
            num_hits = int(np.ceil(hit_fraction * s.num_peptides))

            spot_counts, hit_peptides = simulate_number_hits_per_pool(
                s, num_hits=num_hits
            )
            linear_system = create_linear_system(s, spot_counts)
            result = solve_linear_system(
                linear_system, leave_on_out=False, min_peptide_activity=0.2
            )

            predicted_hits = result.high_confidence_hits

            tp = len(predicted_hits.intersection(hit_peptides))
            fp = len({p for p in predicted_hits if p not in hit_peptides})
            fn = len({p for p in hit_peptides if p not in predicted_hits})

            precision = tp / (tp + fp) if tp + fp > 0 else 0
            recall = tp / (tp + fn) if tp + fn > 0 else 0
            f1 = (
                2 * precision * recall / (precision + recall)
                if precision + recall > 0
                else 0
            )

            ps.append(precision)
            rs.append(recall)
            f1s.append(f1)
    return EvaluationResult(
        precision=np.mean(ps),
        recall=np.mean(rs),
        f1=np.mean(f1s),
        num_pools=s.num_pools(),
        num_violations=count_violations(
            s, error_on_duplicate=False, error_on_extra=False, error_on_missing=False
        ),
    )



from golfy import init, count_violations


def test_valid_init():
    s = init(
        num_peptides=100, max_peptides_per_pool=5, num_replicates=3, strategy="valid"
    )
    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 20
    assert len(s.assignments[0][0]) == 5


def test_singleton_init():
    s = init(
        num_peptides=100,
        max_peptides_per_pool=5,
        num_replicates=3,
        strategy="singleton",
    )
    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 100
    assert len(s.assignments[0][0]) == 1


def test_random_init():
    s = init(
        num_peptides=100, max_peptides_per_pool=5, num_replicates=3, strategy="random"
    )
    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 20
    assert len(s.assignments[0][0]) == 5


def test_greedy_init():
    s = init(
        num_peptides=100, max_peptides_per_pool=5, num_replicates=3, strategy="greedy"
    )
    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 20
    assert len(s.assignments[0][0]) == 5


def test_greedy_init_with_preferred_neighbors():
    """
    Test case from Andy:
    25 peptides / 5 peptides per pool / 3x
        2023-07-06 17:57:30 INFO     peptide_7 and peptide_10
        2023-07-06 17:57:30 INFO     peptide_12 and peptide_17
        2023-07-06 17:57:30 INFO     peptide_14 and peptide_15
        2023-07-06 17:57:30 INFO     peptide_15 and peptide_17
    """
    s = init(
        num_peptides=25,
        max_peptides_per_pool=5,
        num_replicates=3,
        strategy="greedy",
        preferred_neighbors=[(7, 10), (12, 17), (14, 15), (15, 17)],
        verbose=True,
    )
    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 5
    assert len(s.assignments[0][0]) == 5
    # expect that greedy fill of first replicate groups {7, 10} and {12, 17, 14, 15}
    pool_to_peptides = s.assignments[0]
    print(pool_to_peptides)

    peptide_to_pool = s.peptide_to_pool_dict_for_replicate(0)
    print(peptide_to_pool)

    assert peptide_to_pool[7] == peptide_to_pool[10]
    assert peptide_to_pool[12] == peptide_to_pool[17]
    assert peptide_to_pool[14] == peptide_to_pool[15]
    assert peptide_to_pool[15] == peptide_to_pool[17]


def test_repeat_block_init():
    s = init(
        num_peptides=100, max_peptides_per_pool=5, num_replicates=3, strategy="repeat"
    )

    assert len(s.assignments) == 3
    assert len(s.assignments[0]) == 20
    assert len(s.assignments[0][0]) == 5

    assert count_violations(s, error_on_duplicate=False) > 0



from golfy import find_best_design


def test_find_best_design_allow_extra_pools():
    s = find_best_design(100, 20, 3, allow_extra_pools=True)
    assert s.num_peptides == 100
    assert s.num_replicates == 3
    assert s.max_peptides_per_pool == 20
    assert s.num_pools() > 15


def test_find_best_design_no_extra_pools():
    s = find_best_design(100, 20, 3, allow_extra_pools=False)
    assert s.num_peptides == 100
    assert s.num_replicates == 3
    assert s.max_peptides_per_pool == 20
    assert s.num_pools() == 15



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations


from collections import defaultdict
from typing import Iterable, Mapping

from .types import Peptide


def pairs_to_dict(peptide_pairs: Iterable[Peptide]):
    peptide_to_set_dict = defaultdict(set)

    for p1, p2 in peptide_pairs:
        peptide_to_set_dict[p1].add(p2)
        peptide_to_set_dict[p2].add(p1)
    return peptide_to_set_dict


def peptide_to_transitive_neighbors(
    peptide: Peptide, peptide_to_neighbors: Mapping[Peptide, set[Peptide]]
) -> set[Peptide]:
    """
    Find all peptides that are neighbors of a given peptide, either directly or
    indirectly (via a chain of neighbors)
    """
    neighbors = set()
    to_visit = list(peptide_to_neighbors[peptide])
    while to_visit:
        p = to_visit.pop()
        if p not in neighbors and p != peptide:
            neighbors.add(p)
            to_visit.extend(peptide_to_neighbors[p])
    return neighbors


def transitive_closure(
    peptide_to_neighbors: Mapping[Peptide, set[Peptide]]
) -> Mapping[Peptide, set[Peptide]]:
    return {
        peptide: peptide_to_transitive_neighbors(peptide, peptide_to_neighbors)
        for peptide in peptide_to_neighbors
    }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import defaultdict
import random
import time

import numpy as np

from .merging import cleanup
from .design import Design
from .types import SwapCandidateList, ReplicateToNeighborDict, Replicate, Pool, Peptide
from .validity import violations_per_replicate, is_valid, count_violations
from .util import pairs_to_dict


def find_violating_peptides(
    s: Design,
) -> tuple[SwapCandidateList, ReplicateToNeighborDict]:
    replicate_to_pool_to_peptides = s.assignments
    # treat invalid pairs as if they've already been neighbors in a previous round
    peptide_to_neighbors = pairs_to_dict(s.invalid_neighbors)
    needs_swap = []
    replicate_to_neighbor_dict = {}
    for replicate_idx, pool_to_peptides in replicate_to_pool_to_peptides.items():
        for pool_idx, peptides in pool_to_peptides.items():
            for p1 in peptides:
                for p2 in peptides:
                    if p1 != p2:
                        if p2 in peptide_to_neighbors[p1]:
                            needs_swap.append((replicate_idx, pool_idx, p2))
                        else:
                            peptide_to_neighbors[p1].add(p2)
        # neighbor constaints at the end of this replicate
        replicate_to_neighbor_dict[replicate_idx] = {
            peptide: neighbors.copy()
            for (peptide, neighbors) in peptide_to_neighbors.items()
        }
    return needs_swap, replicate_to_neighbor_dict


def _groupby_replicate(
    swap_candidates: SwapCandidateList,
) -> dict[Replicate, tuple[Pool, Peptide]]:
    result = defaultdict(list)
    for replicate_idx, pool_idx, peptide in swap_candidates:
        result[replicate_idx].append((pool_idx, peptide))
    return result


def improve_solution(s: Design, verbose: bool = False):
    replicate_to_pool_to_peptides = s.assignments
    replicate_to_peptide_to_pool = s.replicate_to_peptide_to_pool_dict()

    needs_swap, replicate_to_neighbor_dict = find_violating_peptides(s)

    random.shuffle(needs_swap)

    replicate_to_swap_candidates = _groupby_replicate(needs_swap)

    for replicate_idx in sorted(replicate_to_swap_candidates.keys()):
        pool_to_peptides = replicate_to_pool_to_peptides[replicate_idx]
        peptide_to_pool_idx = replicate_to_peptide_to_pool[replicate_idx]
        neighbors = replicate_to_neighbor_dict[replicate_idx]
        swapped_pools = set()
        swapped_peptides = set()
        empty_pools = {
            pool_idx
            for (pool_idx, pool_peptides) in replicate_to_pool_to_peptides[
                replicate_idx
            ].items()
            if len(pool_peptides) == 0
        }

        all_pool_idx_peptide_pairs = list(pool_to_peptides.items())
        random.shuffle(all_pool_idx_peptide_pairs)
        for pool_idx_a, peptide_a in replicate_to_swap_candidates[replicate_idx]:
            if pool_idx_a in swapped_pools or peptide_a in swapped_peptides:
                continue
            previous_neighbors = neighbors[peptide_a]

            # if a pool is empty, just move the offending peptide there
            if empty_pools:
                pool_idx_b = random.choice(list(empty_pools))
                s.move_peptide(replicate_idx, pool_idx_a, peptide_a, pool_idx_b)
                swapped_pools.add(pool_idx_a)
                swapped_pools.add(pool_idx_b)
                swapped_peptides.add(peptide_a)
                empty_pools.remove(pool_idx_b)

            else:
                other_peptides = []
                for pool_idx_i, pool_peptides_i in all_pool_idx_peptide_pairs:
                    if pool_idx_i == pool_idx_a:
                        continue

                    if pool_idx_i in swapped_pools:
                        continue  # already swapped this pool

                    all_peptides_ok = True
                    for p in pool_peptides_i:
                        if p in swapped_peptides or p in previous_neighbors:
                            all_peptides_ok = False
                            break

                    if all_peptides_ok:
                        other_peptides.extend(
                            [p for p in pool_peptides_i if p not in swapped_peptides]
                        )

                if len(other_peptides) == 0:
                    if verbose:
                        print(
                            "Not able to find a valid peptide to swap with for (%s, %s, %s)"
                            % (replicate_idx, pool_idx_a, peptide_a)
                        )
                    continue

                peptide_b = random.choice(other_peptides)
                pool_idx_b = peptide_to_pool_idx[peptide_b]
                assert peptide_a != peptide_b
                assert pool_idx_a != pool_idx_b
                pool_a = pool_to_peptides[pool_idx_a]
                pool_b = pool_to_peptides[pool_idx_b]
                if (
                    len(pool_a) > 1
                    and len(pool_b) < s.max_peptides_per_pool
                    and random.choice([False, True])
                ):
                    # just move peptide a to the pool with fewer than max peptides
                    if verbose:
                        print(
                            "Moving peptide %d from pool %d to pool %d"
                            % (peptide_a, pool_idx_a, pool_idx_b)
                        )
                    s.move_peptide(replicate_idx, pool_idx_a, peptide_a, pool_idx_b)
                    swapped_pools.add(pool_idx_a)
                    swapped_pools.add(pool_idx_b)
                    swapped_peptides.add(peptide_a)
                else:
                    if verbose:
                        print("Before swap")
                        print(
                            "pool",
                            pool_idx_a,
                            "peptide",
                            peptide_a,
                            pool_to_peptides[pool_idx_a],
                        )
                        print(
                            "pool",
                            pool_idx_b,
                            "peptide",
                            peptide_b,
                            pool_to_peptides[pool_idx_b],
                        )

                    # actually swap them
                    s.swap_peptides(
                        replicate_idx, pool_idx_a, peptide_a, pool_idx_b, peptide_b
                    )

                    if verbose:
                        print("After")
                        print(
                            "pool",
                            pool_idx_a,
                            "peptide",
                            peptide_a,
                            pool_to_peptides[pool_idx_a],
                        )
                        print(
                            "pool",
                            pool_idx_b,
                            "peptide",
                            peptide_b,
                            pool_to_peptides[pool_idx_b],
                        )

                    swapped_pools.add(pool_idx_a)
                    swapped_pools.add(pool_idx_b)
                    swapped_peptides.add(peptide_a)
                    swapped_peptides.add(peptide_b)


def optimize(
    s: Design,
    max_iters: int = 2000,
    verbose: bool = False,
    allow_extra_pools: bool = True,
    return_history: bool = False,
) -> bool:
    """
    Iteratively update solution by randomly swapping a violating peptide with a random other peptide

    Args
    ----
    solution
        Initial solution which will be modified in-place

    max_iters
        Maximum number of swaps to consider performing

    verbose
        print number of violations and pools for each iteration

    allow_extra_pools
        If no improvements have been made for 3 iters, add a pool
        to the last replicate

    return_history
        Return array of constraint validation counts per iteration


    Returns True if non-violating solution found, False if solution still has violations after
    max_iters
    """
    replicate_to_violation_count = violations_per_replicate(s)
    old_num_violations = sum(replicate_to_violation_count.values())

    history = [old_num_violations]
    if verbose:
        print("Initial solution has %s violations" % (old_num_violations,))

    if allow_extra_pools and (old_num_violations / s.num_replicates > 1000):
        # before first iteration, add a lot of empty pools to any replicates with
        # more than 1000 violations
        for replicate_idx, violation_count in replicate_to_violation_count.items():
            if violation_count > 1000:
                num_new_pools = int(np.ceil(violation_count / 1000))
                if verbose:
                    print(
                        "-- replicate %d has %d violations, adding %d pools"
                        % (replicate_idx + 1, violation_count, num_new_pools)
                    )
                for _ in range(num_new_pools):
                    s.add_empty_pool(replicate_idx)

    num_iters_without_improvement = 0
    for i in range(max_iters):
        t0 = time.time()
        improve_solution(s)

        replicate_to_violation_count = violations_per_replicate(s)

        new_num_violations = sum(replicate_to_violation_count.values())

        history.append(new_num_violations)

        if verbose:
            print(
                "%d) %d -> %d violations (%d pools, %0.2fs)"
                % (
                    i + 1,
                    old_num_violations,
                    new_num_violations,
                    s.num_pools(),
                    time.time() - t0,
                )
            )

        if old_num_violations <= new_num_violations:
            num_iters_without_improvement += 1
        else:
            num_iters_without_improvement = 0

        old_num_violations = new_num_violations

        if new_num_violations == 0:
            if verbose:
                print("Found valid solution after %d swaps" % (i + 1,))
            break

        if num_iters_without_improvement >= 3:
            if allow_extra_pools:
                replicates_in_need = [
                    r for (r, v) in replicate_to_violation_count.items() if v > 0
                ]
                assert len(replicates_in_need) > 0
                for replicate_idx in replicates_in_need:
                    if verbose:
                        print(
                            "-- adding pool %d to replicate %d"
                            % (len(s.assignments[replicate_idx]), replicate_idx + 1)
                        )
                    s.add_empty_pool(replicate_idx)
                num_iters_without_improvement = 0
            else:
                # give up if we can't solve this problem
                break

    result = old_num_violations == 0

    # clean up the solution by merging small pools
    # and removing any empty pools
    cleanup(s, verbose=verbose, max_iters=max_iters)

    if result and not is_valid(s):
        raise ValueError(
            "Solution is not valid after cleanup step! Violations: %d"
            % (count_violations(s),)
        )

    if return_history:
        return result, np.array(history)
    else:
        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import math
from typing import Optional


from .design import Design, Spec
from .evaluation import evaluate_design, EvaluationResult
from .initialization import init
from .optimization import optimize
from .types import Replicate, PeptidePairList
from .validity import count_violations


def find_best_design(
    num_peptides: int = 100,
    max_peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools_per_replicate: Optional[int | dict[Replicate, int]] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    """
    Try several different initialization methods and return the solution which
    is optimized to have the fewest violations and the fewest pools.

    Args
    ----
    num_peptides
        The number total number of peptides in the experiment

    max_peptides_per_pool
        Maximum number of peptides which can be in any one pool. This is also the number of peptides
        which will be in most pools if allow_extra_pools is False

    num_replicates
        Number of replicates in the experiment (i.e. how many distinct pools each peptide occurs in)

    num_pools_per_replicate
        Number of pools in each replicate. If None, then this will be set to the minimum number of pools
        required to fit all peptides. If an int, then this will be the number of pools in each replicate.

    invalid_neighbors
        List of peptide pairs which cannot be in the same pool

    preferred_neighbors
        List of peptide pairs which should be in the same pool if possible

    allow_extra_pools
        If True, then the solution can have more than the minimum number of pools required to fit all peptides. If False,
        then the returned solution may not be valid (i.e. some peptide pairs will occur in more than one pool together)

    verbose
        If True, then print out information about the solution as it is being constructed
    """
    shared_kwargs = dict(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        num_pools_per_replicate=num_pools_per_replicate,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        allow_extra_pools=allow_extra_pools,
        verbose=verbose,
    )

    if len(preferred_neighbors) > 0:
        if allow_extra_pools:
            # these are the two methods which can group
            # preferred peptides in pools, either during
            # initialization or during merging
            init_strategies = ["greedy", "singleton"]
        else:
            # if we don't allow extra pools, then we can only
            # group preferred peptides during initialization
            init_strategies = ["greedy"]
    elif allow_extra_pools:
        # if there are no preferred peptides, then we can
        # only use all the initialization methods
        init_strategies = ["greedy", "random", "valid", "singleton"]
    else:
        # only greedy and random initialization methods
        # let us be strict about the number of pools
        init_strategies = ["greedy", "random"]

    designs = {
        strategy: init(strategy=strategy, **shared_kwargs)
        for strategy in init_strategies
    }
    best_design = None
    best_violations = None
    best_num_pools = None

    for strategy, s in designs.items():
        if verbose:
            print(
                "Initialized with strategy '%s': violations=%d, num_pools=%d"
                % (strategy, count_violations(s), s.num_pools())
            )
        optimize(s, allow_extra_pools=allow_extra_pools, verbose=verbose)
        violations = count_violations(s)
        num_pools = s.num_pools()
        if verbose:
            print(
                "-- after optimization of '%s' solution: violations=%d, num_pools=%d"
                % (strategy, violations, num_pools)
            )
        if (
            best_design is None
            or (violations < best_violations)
            or (violations == best_violations and num_pools < best_num_pools)
        ):
            best_design = s
            best_violations = violations
            best_num_pools = num_pools
            if verbose:
                print("^^ new best solution")
    return best_design


def generate_candidate_specs_for_pool_budget(
    num_peptides: int = 100,
    max_pools: int = 96,
    min_num_replicates: int = 2,
    max_num_replicates: int = 5,
    min_max_peptides_per_pool: int = 2,
    max_max_peptides_per_pool: Optional[int] = None,
    allow_extra_pools: Optional[bool] = None,
) -> list[Spec]:
    """
    Generator of candidate configurations from which we might want to try
    to make Designs that fit into the given pool budget.
    """
    if allow_extra_pools is None:
        allow_extra_pools = [True, False]
    elif type(allow_extra_pools) is bool:
        allow_extra_pools = [allow_extra_pools]

    if max_max_peptides_per_pool is None:
        max_max_peptides_per_pool = max(3, num_peptides // 5)

    if min_num_replicates > max_num_replicates:
        max_num_replicates = min_num_replicates

    if min_max_peptides_per_pool > max_max_peptides_per_pool:
        max_max_peptides_per_pool = min_max_peptides_per_pool

    for max_peptides_per_pool in range(
        min_max_peptides_per_pool, max_max_peptides_per_pool
    ):
        if max_peptides_per_pool * max_pools < num_peptides:
            # not enough pools to fit all peptides without replicates
            continue

        for num_replicates in range(min_num_replicates, max_num_replicates + 1):
            min_pools_for_spec = math.ceil(
                num_peptides * num_replicates / max_peptides_per_pool
            )
            if min_pools_for_spec > max_pools:
                # not enough pools to fit all peptides with given number of replicates
                continue

            for curr_allow_extra_pools in allow_extra_pools:
                yield Spec(
                    num_peptides=num_peptides,
                    max_peptides_per_pool=max_peptides_per_pool,
                    num_replicates=num_replicates,
                    allow_extra_pools=curr_allow_extra_pools,
                    invalid_neighbors=[],
                    preferred_neighbors=[],
                )


def score_designs_for_pool_budget(
    num_peptides: int = 100,
    max_pools: int = 96,
    num_simulation_iters: int = 3,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    min_pools: Optional[int] = None,
    min_num_replicates: int = 2,
    max_num_replicates: int = 5,
    min_max_peptides_per_pool: int = 2,
    max_max_peptides_per_pool: Optional[int] = None,
    allow_extra_pools: Optional[bool] = None,
    verbose: bool = False,
) -> list[tuple[Design, EvaluationResult]]:
    assert num_peptides > 1, "No need to pool if there's only one peptide"
    assert max_pools > 1, "Must have more than one pool"
    assert max_pools <= num_peptides, "Can't have more pools than peptides"

    designs = []
    specs = list(
        generate_candidate_specs_for_pool_budget(
            num_peptides=num_peptides,
            max_pools=max_pools,
            min_num_replicates=min_num_replicates,
            max_num_replicates=max_num_replicates,
            min_max_peptides_per_pool=min_max_peptides_per_pool,
            max_max_peptides_per_pool=max_max_peptides_per_pool,
            allow_extra_pools=allow_extra_pools,
        )
    )
    print("Generated %d candidate specs" % (len(specs),))

    shared_kwargs = dict(
        num_peptides=num_peptides,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        verbose=verbose,
    )

    for spec in specs:
        s = find_best_design(
            max_peptides_per_pool=spec.max_peptides_per_pool,
            num_replicates=spec.num_replicates,
            allow_extra_pools=spec.allow_extra_pools,
            **shared_kwargs,
        )
        num_pools = s.num_pools()
        if min_pools is not None and num_pools < min_pools:
            continue

        if num_pools > max_pools:
            print("%s: %d pools > %d max pools" % (spec, num_pools, max_pools))
        else:
            scores = evaluate_design(s, num_simulation_iters)
            print("%s: %s" % (spec, scores))
            designs.append((s, scores))
    return designs


def best_design_for_pool_budget(
    num_peptides: int = 100,
    max_pools: int = 96,
    num_simulation_iters: int = 3,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    min_pools: Optional[int] = None,
    min_num_replicates: int = 3,
    max_num_replicates: int = 8,
    min_max_peptides_per_pool: int = 2,
    max_max_peptides_per_pool: Optional[int] = None,
    verbose: bool = False,
):
    assert num_peptides > 1, "No need to pool if there's only one peptide"
    assert max_pools > 1, "Must have more than one pool"
    assert max_pools <= num_peptides, "Can't have more pools than peptides"
    designs_with_scores = score_designs_for_pool_budget(
        num_peptides=num_peptides,
        max_pools=max_pools,
        num_simulation_iters=num_simulation_iters,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        min_pools=min_pools,
        min_num_replicates=min_num_replicates,
        max_num_replicates=max_num_replicates,
        min_max_peptides_per_pool=min_max_peptides_per_pool,
        max_max_peptides_per_pool=max_max_peptides_per_pool,
        verbose=verbose,
    )
    best_pair = sorted(designs_with_scores, key=lambda x: x[1].sort_key())[0]
    return best_pair[0]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Mapping, Literal
from dataclasses import dataclass

import numpy as np

from .types import Peptide, SpotCounts, Pool, Replicate
from .design import Design


@dataclass
class LinearSystem:
    A: np.ndarray
    b: np.ndarray
    pool_tuple_to_idx: Mapping[tuple[Replicate, Pool], int]
    idx_to_pool_tuple: Mapping[int, tuple[Replicate, Pool]]


def create_linear_system(
    s: Design, spot_counts: SpotCounts, verbose=False
) -> LinearSystem:
    num_peptides = s.num_peptides
    num_pools = s.num_pools()

    A = np.zeros((num_pools, num_peptides + 1)).astype(float)
    b = np.zeros(num_pools).astype(float)

    pool_tuple_to_idx = {}
    idx_to_pool_tuple = {}
    i = 0
    for r, d in spot_counts.items():
        for pool, spots in d.items():
            b[i] = spots
            pool_tuple_to_idx[(r, pool)] = i
            idx_to_pool_tuple[i] = (r, pool)
            for p in s.assignments[r][pool]:
                A[i, p] = 1
            # add a ones column for a constant offset
            A[i, num_peptides] = 1
            i += 1
    if verbose:
        print("Ax = b")
        print("=======")
        print("A.shape: %s" % (A.shape,))
        print("b.shape: %s" % (b.shape,))
        print("A:\n%s" % (A,))
        print("A col sums: %s" % (A.sum(axis=0)))
        print("A row sums: %s" % (A.sum(axis=1)))
        print("b:\n%s" % (b,))
    return LinearSystem(
        A=A,
        b=b,
        pool_tuple_to_idx=pool_tuple_to_idx,
        idx_to_pool_tuple=idx_to_pool_tuple,
    )


@dataclass
class DeconvolutionResult:
    activity_per_peptide: np.ndarray
    prob_hit_per_peptide: np.ndarray
    high_confidence_hits: set[Peptide]
    background: float
    cutoff: float


def solve_linear_system(
    linear_system: LinearSystem,
    min_peptide_activity: float = 1.0,
    leave_on_out=True,
    sparse_solution=True,
    alpha=None,
    verbose=False,
) -> DeconvolutionResult:
    from sklearn.linear_model import Lasso, Ridge

    A = linear_system.A
    b = linear_system.b

    num_pools, num_peptides_with_constant = A.shape
    if alpha is None:
        alpha = 1.0 / (2 * num_peptides_with_constant)
    if verbose:
        print(
            "[solve_linear_system] A.shape: %s, b.shape: %s, alpha = %0.4f"
            % (A.shape, b.shape, alpha)
        )

    num_peptides = num_peptides_with_constant - 1
    row_indices = list(range(num_pools))
    if leave_on_out:
        loo_indices = row_indices
    else:
        loo_indices = [None]

    avg_activity = np.zeros(num_peptides)
    frac_hit = np.zeros(num_peptides)
    cutoff = 0
    for loo_idx in loo_indices:
        subset_indices = np.array([i for i in row_indices if i != loo_idx])
        A_subset = A[subset_indices, :]
        b_subset = b[subset_indices]
        b_min = np.percentile(b_subset, 1)
        b_max = np.percentile(b_subset, 99)
        scale = b_max - b_min

        assert scale > 0

        if sparse_solution:
            # L1 minimization to get a small set of confident active peptides
            model = Lasso(
                fit_intercept=False,
                positive=True,
                alpha=alpha,
                selection="random",
            )

        else:
            # this will work horribly, have fun
            model = Ridge(fit_intercept=False, positive=True, alpha=alpha)

        model.fit(A_subset, np.maximum(0, b_subset - b_min) / scale)
        x_with_offset = scale * model.coef_

        x, c = x_with_offset[:-1], x_with_offset[-1]

        if verbose:
            print("x = %s" % (x,))
            print("c = %s" % (c,))

        avg_activity += x
        curr_cutoff = choose_cutoff(
            A=A_subset,
            b=b_subset,
            estimated_constant_offsset=c,
            min_peptide_activity=min_peptide_activity,
            sparse_estimate=sparse_solution,
        )

        frac_hit += (x > curr_cutoff).astype(float)
        cutoff += curr_cutoff

    avg_activity /= len(loo_indices)
    frac_hit /= len(loo_indices)
    high_confidence_hits = set(np.where(frac_hit > 0.5)[0])
    cutoff /= len(loo_indices)

    return DeconvolutionResult(
        activity_per_peptide=avg_activity,
        prob_hit_per_peptide=frac_hit,
        high_confidence_hits=high_confidence_hits,
        background=c,
        cutoff=cutoff,
    )


def em_step(
    linear_system: LinearSystem,
    beta: np.ndarray,
    # stability_factor=1e-6,
) -> np.ndarray:
    """
    Implements equation 5 from Strom et al (2016)
        (X.T @ (y / (X @ beta))) * (beta / X.sum(0))
    where X is the design matrix, y is the pool activities, and beta is the current estimate of the peptide activities
    """
    X = linear_system.A
    y = linear_system.b
    return (X.T @ (y / (X @ beta))) * (beta / X.sum(0))


def em_init(linear_system: LinearSystem) -> np.ndarray:
    """
    Initialize beta for E-M algorithm from Strom et al (2016)
    to have each peptide be the mean of its pool activities
    """
    X = linear_system.A
    y = linear_system.b
    return X.T @ y / X.sum(0)


def choose_cutoff(
    A: np.ndarray,
    b: np.ndarray,
    estimated_constant_offsset: float,
    min_peptide_activity: float,
    sparse_estimate: bool,
) -> float:
    empty_wells = np.where(A.sum(1) == 0)[0]
    any_empty_wells = len(empty_wells) > 0
    if any_empty_wells:
        empty_well_background = b[empty_wells].mean()
    else:
        empty_well_background = 0

    estimated_constant_offsset = max(0, estimated_constant_offsset)
    print(A, b, estimated_constant_offsset, empty_well_background, min_peptide_activity)
    if sparse_estimate:
        return (
            np.sqrt(estimated_constant_offsset)
            + empty_well_background
            + min_peptide_activity
        )
    else:
        return (
            3 * estimated_constant_offsset
            + empty_well_background
            + min_peptide_activity
        )


def em_deconvolve(
    linear_system: LinearSystem,
    min_peptide_activity=1.0,
    max_iters=500,
    verbose=False,
    tol=1e-3,
    sparsity_threshold=1e-4,
) -> DeconvolutionResult:
    """
    Implements the expectation-maximization algorithm from
    "A statistical approach to determining responses to individual peptides from pooled-peptide ELISpot data"
    """
    beta = em_init(linear_system=linear_system)
    last_beta = beta.copy()
    for i in range(max_iters):
        beta = em_step(linear_system=linear_system, beta=beta)
        mean_abs_diff = np.mean(np.abs(beta - last_beta))
        if verbose:
            print(
                "[em_deconvolve] Iter %d: beta = %s (mean abs change=%f)"
                % (i, beta.round(2), mean_abs_diff)
            )
        if mean_abs_diff < tol:
            if verbose:
                print("[em_deconvolve] Stopping at iter %d" % i)
            break
        beta[np.abs(beta) < sparsity_threshold] = 0
        last_beta = beta.copy()
    activity_per_peptide = beta[:-1]
    activity_per_peptide[activity_per_peptide < sparsity_threshold] = 0
    print(activity_per_peptide.round(1))
    background = beta[-1]
    cutoff = choose_cutoff(
        A=linear_system.A,
        b=linear_system.b,
        estimated_constant_offsset=background,
        min_peptide_activity=min_peptide_activity,
        sparse_estimate=False,
    )
    above_limit = activity_per_peptide > cutoff
    return DeconvolutionResult(
        activity_per_peptide=activity_per_peptide,
        prob_hit_per_peptide=above_limit.astype(float),
        high_confidence_hits=set(np.where(above_limit)[0]),
        background=background,
        cutoff=cutoff,
    )


def deconvolve(
    s: Design,
    spot_counts: SpotCounts,
    method: Literal["lasso", "ridge", "em"] = "lasso",
    min_peptide_activity=1.0,
    verbose=False,
):
    """
    Arguments
    ---------
    s
        Dictionary containing mapping of replicates -> pools -> peptides

    spot_counts
        Dictionary containing mapping of replicates -> pool -> spot counts

    min_peptide_activity
        Minimum estimated activity of a peptide to be considered for hit set

    method
        - "lasso": L1 regularized linear regression
        - "ridge": L2 regularized linear regression
        - "em": EM algorithm from Strom et al. 2016
                      "A statistical approach to determining responses to individual peptides from pooled-peptide ELISpot data"

    verbose
        Print diagnostic information

    Returns DeconvolutionResult
    """
    linear_system = create_linear_system(s, spot_counts, verbose=verbose)
    if method == "lasso":
        return solve_linear_system(
            linear_system,
            sparse_solution=True,
            leave_on_out=False,
            min_peptide_activity=min_peptide_activity,
        )
    elif method == "ridge":
        return solve_linear_system(
            linear_system,
            sparse_solution=False,
            leave_on_out=False,
            min_peptide_activity=min_peptide_activity,
        )
    elif method == "em":
        return em_deconvolve(
            linear_system, min_peptide_activity=min_peptide_activity, verbose=verbose
        )
    else:
        raise ValueError("Unknown method: %s" % (method,))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import defaultdict
import math
from typing import Optional, Literal

import numpy as np

from .design import Design
from .types import PeptidePairList, Replicate, Peptide
from .util import pairs_to_dict, transitive_closure
from .validity import count_violations


def _pools_per_replicate(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    allow_extra_pools: bool,
    verbose: bool,
) -> dict[Replicate, int]:
    too_few_pools = num_peptides < max_peptides_per_pool**2
    if too_few_pools and not allow_extra_pools and verbose:
        print(
            "Warning: cannot satisfy constraints with %d peptides and %d peptides per pool (max allowed %d)"
            % (num_peptides, max_peptides_per_pool, int(np.sqrt(num_peptides)))
        )
    num_pools_first_replicate = int(np.ceil(num_peptides / max_peptides_per_pool))
    result = {0: num_pools_first_replicate}
    for replicate_idx in range(1, num_replicates):
        if too_few_pools and allow_extra_pools:
            max_peptides_per_pool = num_pools_first_replicate
            result[replicate_idx] = int(np.ceil(num_peptides / max_peptides_per_pool))
        else:
            result[replicate_idx] = num_pools_first_replicate
    return result


def _repeat_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
):
    """
    Assign peptides to random blocks and then repeat those blocks r times
    """
    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        peptide_array = np.arange(num_peptides)
        np.random.shuffle(peptide_array)
        pool_assignments = {}
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(np.ceil(num_peptides / num_pools))
        for j in range(num_pools):
            start_idx = peptides_per_pool * j
            end_idx = peptides_per_pool * (j + 1)
            pool_assignments[j] = peptide_array[start_idx:end_idx]
        replicate_to_pool_to_peptides[i] = pool_assignments
    return Design(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
        allow_extra_pools=allow_extra_pools,
    )


def _random_peptide_order(
    num_peptides: int, peptide_to_preferred: dict[Peptide, set[Peptide]]
):
    """
    Return the peptide indices in a random order but make sure the ones with preferred neighbors are first
    """
    random_peptide_order = np.arange(num_peptides)
    np.random.shuffle(random_peptide_order)
    assert len(set(random_peptide_order)) == num_peptides
    # assign all peptides with preferred neighbors first
    peptides_with_preferred_neighbors = [
        p for p in random_peptide_order if peptide_to_preferred.get(p)
    ]
    peptides_without_preferred_neighbors = [
        p for p in random_peptide_order if not peptide_to_preferred.get(p)
    ]
    peptide_list = (
        peptides_with_preferred_neighbors + peptides_without_preferred_neighbors
    )
    return peptide_list


def _random_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        peptide_array = np.arange(num_peptides)
        np.random.shuffle(peptide_array)
        pool_assignments = {}
        replicate_to_pool_to_peptides[i] = pool_assignments
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(np.ceil(num_peptides / num_pools))
        for j in range(num_pools):
            start_idx = peptides_per_pool * j
            end_idx = peptides_per_pool * (j + 1)
            pool_assignments[j] = peptide_array[start_idx:end_idx]

    return Design(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
        allow_extra_pools=allow_extra_pools,
    )


def _singleton_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    """
    Initialize every peptide to be in its own pool
    """

    replicate_to_pool_to_peptides = {}
    for i in range(num_replicates):
        pool_to_peptides = {}
        for p in range(num_peptides):
            pool_to_peptides[p] = np.array([p])
        replicate_to_pool_to_peptides[i] = pool_to_peptides
    return Design(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
        allow_extra_pools=allow_extra_pools,
    )


def _valid_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    peptide_to_invalid = pairs_to_dict(invalid_neighbors)
    peptide_to_preferred = transitive_closure(pairs_to_dict(preferred_neighbors))
    if verbose:
        print("[_valid_init] Invalid neighbors: %s" % (peptide_to_invalid,))
        print("[_valid_init] Preferred neighbors: %s" % (peptide_to_preferred,))

    replicate_to_pool_to_peptides = {}

    for i in range(num_replicates):
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(math.ceil(num_peptides / num_pools))

        peptide_to_pool = {}
        pool_to_peptides = defaultdict(set)

        def add_to_pool(peptide, pool_idx):
            pool = pool_to_peptides[pool_idx]
            pool.add(peptide)
            peptide_to_pool[peptide] = pool_idx
            for other_peptide in pool:
                peptide_to_invalid[peptide].add(other_peptide)
                peptide_to_invalid[other_peptide].add(peptide)

        def curr_num_pools():
            return len(pool_to_peptides)

        def make_new_pool(peptide):
            new_pool_idx = curr_num_pools()
            add_to_pool(peptide, new_pool_idx)
            return new_pool_idx

        for peptide in _random_peptide_order(
            num_peptides=num_peptides, peptide_to_preferred=peptide_to_preferred
        ):
            for preferred_neighbor in peptide_to_preferred.get(peptide, []):
                if preferred_neighbor in peptide_to_invalid[peptide]:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d already invalid with preferred neighbor %d"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool_idx = peptide_to_pool.get(preferred_neighbor)
                if preferred_pool_idx is None:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d, preferred neighbor %d not in a pool yet"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool = pool_to_peptides[preferred_pool_idx]
                if len(preferred_pool) >= peptides_per_pool:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, preferred neighbor %d in pool %d which is already full"
                            % (i, preferred_neighbor, preferred_pool_idx)
                        )
                    continue

                add_to_pool(peptide, preferred_pool_idx)
                if verbose:
                    print(
                        "[_greedy_init] replicate %d, adding peptide %d to preferred pool %d to pair with peptide %d"
                        % (i, peptide, preferred_pool_idx, preferred_neighbor)
                    )
                break
            # if we didn't get a preferred peptide pool that's valid
            # and there's room for more pools, just make a singleton
            if peptide not in peptide_to_pool:
                if curr_num_pools() < num_pools:
                    assigned_pool_idx = make_new_pool(peptide)
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, making new pool %d for peptide %d"
                            % (i, assigned_pool_idx, peptide)
                        )

            # otherwise, try to find a valid pool
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        disallowed_peptides = peptide_to_invalid[peptide]
                        valid = all(
                            [
                                other_peptide not in disallowed_peptides
                                for other_peptide in pool
                            ]
                        )
                        if valid:
                            add_to_pool(peptide, pool_idx)
                            break

            # otherwise, make a new pool
            if peptide not in peptide_to_pool:
                if not allow_extra_pools and verbose:
                    print(
                        "Warning: unable to create valid configuration without exceeding max pools for replicate %d (max %d, adding %d)"
                        % (i + 1, num_pools, curr_num_pools() + 1)
                    )
                make_new_pool(peptide)

        replicate_to_pool_to_peptides[i] = {
            pool_idx: np.array(sorted(peptides))
            for (pool_idx, peptides) in pool_to_peptides.items()
        }
    return Design(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
        allow_extra_pools=allow_extra_pools,
    )


def _greedy_init(
    num_peptides: int,
    max_peptides_per_pool: int,
    num_replicates: int,
    num_pools_per_replicate: dict[Replicate, int],
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    peptide_to_invalid = pairs_to_dict(invalid_neighbors)
    peptide_to_preferred = transitive_closure(pairs_to_dict(preferred_neighbors))
    if verbose:
        print("[_greedy_init] Invalid neighbors: %s" % (peptide_to_invalid,))
        print("[_greedy_init] Preferred neighbors: %s" % (peptide_to_preferred,))

    replicate_to_pool_to_peptides = {}

    for i in range(num_replicates):
        num_pools = num_pools_per_replicate[i]
        peptides_per_pool = int(math.ceil(num_peptides / num_pools))

        peptide_to_pool = {}
        pool_to_peptides = defaultdict(set)

        def add_to_pool(peptide, pool_idx):
            pool = pool_to_peptides[pool_idx]
            pool.add(peptide)
            peptide_to_pool[peptide] = pool_idx
            for other_peptide in pool:
                peptide_to_invalid[peptide].add(other_peptide)
                peptide_to_invalid[other_peptide].add(peptide)

        def curr_num_pools():
            return len(pool_to_peptides)

        def make_new_pool(peptide):
            new_pool_idx = curr_num_pools()
            add_to_pool(peptide, new_pool_idx)
            return new_pool_idx

        for peptide in _random_peptide_order(
            num_peptides=num_peptides, peptide_to_preferred=peptide_to_preferred
        ):
            for preferred_neighbor in peptide_to_preferred.get(peptide, []):
                if preferred_neighbor in peptide_to_invalid[peptide]:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d already invalid with preferred neighbor %d"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool_idx = peptide_to_pool.get(preferred_neighbor)
                if preferred_pool_idx is None:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, peptide %d, preferred neighbor %d not in a pool yet"
                            % (i, peptide, preferred_neighbor)
                        )
                    continue
                preferred_pool = pool_to_peptides[preferred_pool_idx]
                if len(preferred_pool) >= peptides_per_pool:
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, preferred neighbor %d in pool %d which is already full"
                            % (i, preferred_neighbor, preferred_pool_idx)
                        )
                    continue

                add_to_pool(peptide, preferred_pool_idx)
                if verbose:
                    print(
                        "[_greedy_init] replicate %d, adding peptide %d to preferred pool %d to pair with peptide %d"
                        % (i, peptide, preferred_pool_idx, preferred_neighbor)
                    )
                break
            # if we didn't get a preferred peptide pool that's valid
            # and there's room for more pools, just make a singleton
            if peptide not in peptide_to_pool:
                if curr_num_pools() < num_pools:
                    assigned_pool_idx = make_new_pool(peptide)
                    if verbose:
                        print(
                            "[_greedy_init] replicate %d, making new pool %d for peptide %d"
                            % (i, assigned_pool_idx, peptide)
                        )

            # otherwise, try to find a valid pool
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        disallowed_peptides = peptide_to_invalid[peptide]
                        valid = all(
                            [
                                other_peptide not in disallowed_peptides
                                for other_peptide in pool
                            ]
                        )
                        if valid:
                            add_to_pool(peptide, pool_idx)
                            break

            # otherwise, pick any pool less than the max size
            if peptide not in peptide_to_pool:
                for pool_idx, pool in pool_to_peptides.items():
                    if len(pool) < peptides_per_pool:
                        add_to_pool(peptide, pool_idx)
                        break

            # lastly, violate the num_pools constraint to make a
            # new singleton anyways
            if peptide not in peptide_to_pool:
                if allow_extra_pools:
                    make_new_pool(peptide)
                else:
                    raise ValueError("Unable to create valid configuration")

        replicate_to_pool_to_peptides[i] = {
            pool_idx: np.array(sorted(peptides))
            for (pool_idx, peptides) in pool_to_peptides.items()
        }

    return Design(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        assignments=replicate_to_pool_to_peptides,
        allow_extra_pools=allow_extra_pools,
    )


def init(
    num_peptides: int = 100,
    max_peptides_per_pool: int = 5,
    num_replicates: int = 3,
    num_pools_per_replicate: Optional[int | dict[Replicate, int]] = None,
    invalid_neighbors: PeptidePairList = [],
    preferred_neighbors: PeptidePairList = [],
    strategy: Literal["greedy", "random", "valid", "singleton", "repeat"] = "greedy",
    allow_extra_pools: bool = False,
    verbose: bool = False,
) -> Design:
    """
    Initialize a Solution for a given configuration

    Args
    ----
    num_peptides
        number of peptides (default: 100)

    peptides_per_pool
        number of peptides per pool (default: 5)

    num_replicates
        number of replicates in the Solution (default: 3)

    num_pools_per_replicate
        number of pools in each replicate (default: ceil(num_peptides / peptides_per_pool))

    invalid_neighbors
        list of peptide pairs that cannot be in the same pool

    preferred_neighbors
        list of peptide pairs that should be in the same pool

    strategy
        initialization strategy, one of {"greedy", "random", "singleton", "valid", "repeat"} (default: "greedy")

    allow_extra_pools
        whether to allow extra pools to be created to satisfy constraints (default: False)

    verbose
        print diagnostic information during initialization (default: False)
    """
    fn = {
        "greedy": _greedy_init,
        "random": _random_init,
        "singleton": _singleton_init,
        "valid": _valid_init,
        "repeat": _repeat_init,
    }.get(strategy)
    if num_pools_per_replicate is None:
        num_pools_per_replicate = _pools_per_replicate(
            num_peptides=num_peptides,
            max_peptides_per_pool=max_peptides_per_pool,
            num_replicates=num_replicates,
            allow_extra_pools=allow_extra_pools,
            verbose=verbose,
        )
    elif isinstance(num_pools_per_replicate, int):
        num_pools_per_replicate = {
            i: num_pools_per_replicate for i in range(num_replicates)
        }
    if fn is None:
        raise ValueError("Unknown initialization strategy: '%s'" % (strategy,))

    if verbose:
        print(
            "Using '%s' initialization strategy, pools per replicate: %s"
            % (strategy, num_pools_per_replicate)
        )
    kwargs = dict(
        num_peptides=num_peptides,
        max_peptides_per_pool=max_peptides_per_pool,
        num_replicates=num_replicates,
        num_pools_per_replicate=num_pools_per_replicate,
        invalid_neighbors=invalid_neighbors,
        preferred_neighbors=preferred_neighbors,
        allow_extra_pools=allow_extra_pools,
        verbose=verbose,
    )
    s = fn(**kwargs)
    violations = count_violations(s)
    if verbose:
        print("Generated solution with %d initial violations" % (violations))
    return s



from golfy import init, optimize, is_valid
from golfy.simulation import simulate_elispot_counts
from golfy.deconvolution import (
    create_linear_system,
    solve_linear_system,
    deconvolve,
)


def test_deconvolution_lasso():
    s = init(100, 3, 5)
    optimize(s)
    assert is_valid(s)
    # make sure the margin between background and hit activities
    # is so large that we never falsely identify a background peptide
    counts, hit_peptides = simulate_elispot_counts(
        s,
        num_hits=3,
        max_background_peptide_activity=1,
        min_hit_peptide_activity=100,
    )

    linear_system = create_linear_system(s, counts)
    for loo in [False, True]:
        # solver should work regardless of whether we use LOO to estimate
        # probabilities of hits or not
        solution = solve_linear_system(linear_system, leave_on_out=loo)
        assert solution.high_confidence_hits == hit_peptides


def test_deconvolution_em():
    s = init(100, 3, 5)
    optimize(s)
    assert is_valid(s)
    # make sure the margin between background and hit activities
    # is so large that we never falsely identify a background peptide
    counts, hit_peptides = simulate_elispot_counts(
        s,
        num_hits=3,
        max_background_peptide_activity=1,
        min_hit_peptide_activity=100,
    )

    solution = deconvolve(s, counts, method="em")

    assert solution.high_confidence_hits.issuperset(hit_peptides)


if __name__ == "__main__":
    d = globals().copy()
    for name in d:
        if name.startswith("test_"):
            print("Running %s" % (name,))
            obj = d[name]
            obj()



from golfy import best_design_for_pool_budget


def test_best_design_for_pool_budget():
    s = best_design_for_pool_budget(num_peptides=50, max_pools=20)
    assert s.num_peptides == 50
    assert s.num_replicates > 1
    assert s.num_pools() <= 20



from golfy.simulation import (
    simulate_elispot_counts,
    simulate_any_hits_per_pool,
    simulate_number_hits_per_pool,
)
from golfy import init


def test_simulate_elispot_counts_known_hits():
    # create 100 blocks of one peptide each
    s = init(20, 1, 5, strategy="singleton")
    # make the first two peptides hits
    hits = {0, 1}
    spot_counts, hits2 = simulate_elispot_counts(s, hit_peptides=hits)
    assert hits == hits2


def test_simulate_any_hits_per_pool_known_hits():
    # create 100 blocks of one peptide each
    s = init(20, 1, 5, strategy="singleton")
    hits = {0, 1}
    spot_counts, hits2 = simulate_any_hits_per_pool(s, hit_peptides=hits)
    assert hits == hits2


def test_simulate_number_hits_per_pool_known_hits():
    # create 100 blocks of one peptide each
    s = init(20, 1, 5, strategy="singleton")
    hits = {0, 1}
    spot_counts, hits2 = simulate_number_hits_per_pool(s, hit_peptides=hits)
    assert hits == hits2



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Iterable, Mapping

import numpy as np

from .types import PeptidePairList, Replicate, Pool, Peptide


@dataclass
class Spec:
    num_peptides: int
    max_peptides_per_pool: int
    num_replicates: int
    allow_extra_pools: bool
    invalid_neighbors: PeptidePairList
    preferred_neighbors: PeptidePairList

    def to_tuple(self):
        return (
            self.num_peptides,
            self.max_peptides_per_pool,
            self.num_replicates,
            self.allow_extra_pools,
            tuple(sorted(self.invalid_neighbors)),
            tuple(sorted(self.preferred_neighbors)),
        )

    def __eq__(self, other):
        return isinstance(other, Spec) and self.to_tuple() == other.to_tuple()

    def __hash__(self):
        return hash(self.to_tuple())


@dataclass
class Design(Spec):
    assignments: Mapping[Replicate, Mapping[Pool, Iterable[Peptide]]]

    def to_spec(self) -> Spec:
        return Spec(
            num_peptides=self.num_peptides,
            max_peptides_per_pool=self.max_peptides_per_pool,
            num_replicates=self.num_replicates,
            allow_extra_pools=self.allow_extra_pools,
            invalid_neighbors=self.invalid_neighbors,
            preferred_neighbors=self.preferred_neighbors,
        )

    def move_peptide(
        self,
        replicate_idx: Replicate,
        old_pool_idx: Pool,
        peptide: Peptide,
        new_pool_idx: Pool,
    ):
        """
        Move a peptide from its current pool to a new pool
        """
        pool_to_peptides = self.assignments[replicate_idx]
        old_pool = pool_to_peptides[old_pool_idx]
        new_pool = pool_to_peptides[new_pool_idx]
        assert len(new_pool) < self.max_peptides_per_pool
        pool_to_peptides[new_pool_idx] = np.array(list(new_pool) + [peptide])

        pool_to_peptides[old_pool_idx] = np.array([p for p in old_pool if p != peptide])

    def swap_peptides(
        self,
        replicate_idx: Replicate,
        pool_idx_a: Pool,
        peptide_a: Peptide,
        pool_idx_b: Pool,
        peptide_b: Peptide,
    ):
        """
        Move peptide A from its current pool to the pool of peptide B (and vice versa)
        """
        pool_to_peptides = self.assignments[replicate_idx]
        pool_a = pool_to_peptides[pool_idx_a]
        pool_b = pool_to_peptides[pool_idx_b]
        pool_to_peptides[pool_idx_a] = np.array(
            [i for i in pool_a if i != peptide_a] + [peptide_b]
        )
        pool_to_peptides[pool_idx_b] = np.array(
            [i for i in pool_b if i != peptide_b] + [peptide_a]
        )

    def add_empty_pool(self, replicate_idx: int):
        """
        Add an empty pool to the given replicate
        """
        replicate = self.assignments[replicate_idx]
        num_pools = len(replicate)
        replicate[num_pools] = np.array([])

    def remove_empty_pools(self):
        """
        Delete any empty pools and renumber the pools to be contiguous
        """
        for replicate_idx, pool_to_peptides in self.assignments.items():
            to_delete = [
                pool_idx
                for (pool_idx, pool) in pool_to_peptides.items()
                if len(pool) == 0
            ]
            if len(to_delete) > 0:
                for pool_idx in to_delete:
                    del pool_to_peptides[pool_idx]
                index_mapping = {
                    old_idx: new_idx
                    for (new_idx, old_idx) in enumerate(sorted(pool_to_peptides.keys()))
                }
                self.assignments[replicate_idx] = {
                    index_mapping[pool_idx]: pool
                    for (pool_idx, pool) in pool_to_peptides.items()
                }

    def num_pools(self):
        total = 0
        for pool_to_peptides in self.assignments.values():
            total += len(pool_to_peptides)
        return total

    def average_peptides_per_pool(self):
        return self.num_peptides * self.num_replicates / self.num_pools()

    def peptide_to_pool_dict_for_replicate(
        self, replicate_idx: Replicate
    ) -> Mapping[Peptide, Pool]:
        return {
            p: pool
            for (pool, peptides) in self.assignments[replicate_idx].items()
            for p in peptides
        }

    def replicate_to_peptide_to_pool_dict(
        self,
    ) -> Mapping[Replicate, Mapping[Peptide, Pool]]:
        return {
            replicate_idx: self.peptide_to_pool_dict_for_replicate(replicate_idx)
            for replicate_idx in range(self.num_replicates)
        }

    def peptide_to_replicate_pool_pairs_dict(
        self,
    ) -> Mapping[Peptide, list[tuple[Replicate, Pool]]]:
        result = defaultdict(list)
        for (
            replicate_idx,
            peptide_to_pool,
        ) in self.replicate_to_peptide_to_pool_dict().items():
            for peptide, pool in peptide_to_pool.items():
                result[peptide].append((replicate_idx, pool))
        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional

import numpy as np

from .design import Design
from .types import Peptide, SpotCounts


def _split_peptides_by_hits(
    num_peptides: int,
    num_hits: int,
    hit_peptides: Optional[set[Peptide]] = None,
    verbose: bool = False,
) -> tuple[set[Peptide], set[Peptide]]:
    if hit_peptides is None:
        all_peptides = np.arange(num_peptides)
        np.random.shuffle(all_peptides)
        hit_peptides = set(all_peptides[:num_hits])
        not_hit_peptides = set(all_peptides[num_hits:])
        if verbose:
            print("Hits: %s" % (hit_peptides,))
    else:
        hit_peptides = set(hit_peptides)
        not_hit_peptides = set(range(num_peptides)).difference(hit_peptides)
    return (hit_peptides, not_hit_peptides)


def simulate_number_hits_per_pool(
    s: Design,
    num_hits: int = 5,
    activity_level: int = 1,
    hit_peptides: Optional[set[Peptide]] = None,
    verbose: bool = False,
) -> tuple[SpotCounts, set[Peptide]]:
    """
    Returns spot counts corresponding to the number of hit peptides per pool and set of hit peptides.
    """
    n = s.num_peptides
    hit_peptides, _ = _split_peptides_by_hits(
        n, num_hits, hit_peptides=hit_peptides, verbose=verbose
    )
    activity = np.zeros(n)
    activity[np.array(list(hit_peptides))] = activity_level
    spot_counts: SpotCounts = {
        r: {pool: sum(activity[peptides]) for (pool, peptides) in d.items()}
        for (r, d) in s.assignments.items()
    }
    if verbose:
        print("Spot counts: %s" % (spot_counts,))
    return spot_counts, hit_peptides


def simulate_any_hits_per_pool(
    s: Design,
    num_hits: int = 5,
    hit_peptides: Optional[set[Peptide]] = None,
    verbose: bool = False,
) -> tuple[SpotCounts, set[Peptide]]:
    """
    Returns a spot counts dictionary with a 1 if the pool contained a hit peptides and a 0 otherwise.
    Also returns the
    """
    (spot_counts, hit_peptides) = simulate_number_hits_per_pool(
        s, num_hits, 1, hit_peptides=hit_peptides, verbose=verbose
    )
    spot_counts_binary = {
        r: {pool: 1 if spots > 0 else 0 for (pool, spots) in d.items()}
        for (r, d) in spot_counts.items()
    }
    return (spot_counts_binary, hit_peptides)


def simulate_elispot_counts(
    s: Design,
    num_hits=5,
    max_activity_per_well=2000,
    min_background_peptide_activity=0,
    max_background_peptide_activity=20,
    min_hit_peptide_activity=50,
    max_hit_peptide_activity=500,
    hit_peptides: Optional[set[Peptide]] = None,
    verbose=False,
) -> tuple[SpotCounts, set[Peptide]]:
    num_peptides = s.num_peptides

    background_peptide_activity_range = (
        max_background_peptide_activity - min_background_peptide_activity
    )
    hit_peptide_activity_range = max_hit_peptide_activity - min_hit_peptide_activity

    hit_peptides, not_hit_peptides = _split_peptides_by_hits(
        s.num_peptides, num_hits, hit_peptides=hit_peptides, verbose=verbose
    )

    background = (
        np.random.rand(num_peptides) * background_peptide_activity_range
        + min_background_peptide_activity
    )
    if verbose:
        print("Background activity: %s" % (background,))

    hit_activity = (
        np.random.rand(num_peptides) * hit_peptide_activity_range
        + min_hit_peptide_activity
    )
    hit_activity[np.array(list(not_hit_peptides))] = 0
    if verbose:
        print("Hit activity: %s" % (hit_activity,))

    spot_counts: SpotCounts = {
        r: {
            pool: min(
                max_activity_per_well,
                int(sum([background[i] + hit_activity[i] for i in peptides])),
            )
            for (pool, peptides) in d.items()
        }
        for (r, d) in s.assignments.items()
    }
    if verbose:
        print("Spot counts: %s" % (spot_counts,))
    return (spot_counts, hit_peptides)



from golfy import init, optimize, is_valid


def test_optimize():
    s = init(num_peptides=200, max_peptides_per_pool=10, num_replicates=3)
    assert not is_valid(s)
    optimize(s)
    assert is_valid(s)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from .design import Design
from .types import Replicate,  Mapping
from .util import pairs_to_dict


def violations_per_replicate(
    s: Design,
    error_on_duplicate=True,
    error_on_extra=True,
    error_on_missing=True,
    verbose=False,
) -> Mapping[Replicate, int]:
    replicate_to_violations = {}

    # treat invalid pairs as if they've already been neighbors in a previous round
    peptide_to_neighbors = pairs_to_dict(s.invalid_neighbors)

    for replicate_idx, pool_to_peptides in s.assignments.items():
        violations = 0
        replicate_num = replicate_idx + 1

        # first check each peptide occurs once per replicate
        seen_peptides = set()
        for peptides in pool_to_peptides.values():
            for p in peptides:
                if p in seen_peptides:
                    if error_on_duplicate:
                        raise ValueError(
                            "Peptide %s twice in replicate %s" % (p, replicate_num)
                        )
                    violations += 1
                seen_peptides.add(p)
        expected_peptides = set(range(s.num_peptides))
        if seen_peptides != expected_peptides:
            extra_peptides = seen_peptides.difference(expected_peptides)
            if extra_peptides:
                if error_on_extra:
                    raise ValueError(
                        "Unexpected extra peptides in replicate %d: %s"
                        % (
                            replicate_num,
                            extra_peptides,
                        )
                    )
                violations += len(extra_peptides)
            missing_peptides = expected_peptides.difference(seen_peptides)
            if missing_peptides:
                if error_on_missing:
                    raise ValueError(
                        "Unexpected missing peptides in replicate %d: %s"
                        % (replicate_num, missing_peptides)
                    )
                violations += len(missing_peptides)

        # next check to make sure that each peptides only paired with another at most once
        for peptides in pool_to_peptides.values():
            for p1 in peptides:
                for p2 in peptides:
                    if p1 != p2:
                        if p2 in peptide_to_neighbors[p1]:
                            if verbose:
                                print(
                                    "Peptides %s and %s already together previous pool before replicate %s"
                                    % (p1, p2, replicate_num)
                                )
                            violations += 1
                        peptide_to_neighbors[p1].add(p2)
        replicate_to_violations[replicate_idx] = violations
    return replicate_to_violations


def count_violations(
    s: Design,
    error_on_duplicate=True,
    error_on_extra=True,
    error_on_missing=True,
    verbose=False,
) -> int:
    return sum(
        violations_per_replicate(
            s,
            error_on_duplicate=error_on_duplicate,
            error_on_extra=error_on_extra,
            error_on_missing=error_on_missing,
            verbose=verbose,
        ).values()
    )


def is_valid(
    s: Design,
    error_on_duplicate=True,
    error_on_extra=True,
    error_on_missing=True,
    verbose=False,
) -> bool:
    return (
        count_violations(
            s,
            error_on_duplicate=error_on_duplicate,
            error_on_extra=error_on_extra,
            error_on_missing=error_on_missing,
            verbose=verbose,
        )
        == 0
    )
